// SaslSspi.cpp : Defines the entry point for the application.
//

#include "net_za_darkskies_security_sasl_SSPIImpl.h"
#include "precomp.h"
//#include "transport.h"
#include "model.h"

//Model*		s_pModel;  // SSPI model
//wchar_t		szSPN[256];

static jclass clsSSPIImpl;
static jfieldID fidModel;
static jfieldID fidAuthorizationId;
static jfieldID fidCompleted;

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_initIDs
  (JNIEnv * env, jclass jcls)
{
	clsSSPIImpl = jcls;
	fidModel = env->GetFieldID(jcls, "model", "J");
	fidAuthorizationId = env->GetFieldID(jcls, "authorizationId", "Ljava/lang/String;");
	fidCompleted = env->GetFieldID(jcls, "completed", "Z");

}

void _err(JNIEnv * env, const wchar_t* psz, bool bRecoverable = false, DWORD err = GetLastError()) {
	static jclass clsExc = env->FindClass("javax/security/sasl/SaslException");

	wchar_t szErr[256];
	if (!FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		0, err, 0, szErr, sizeof szErr / sizeof *szErr, 0))
		wsprintf(szErr, L"Unknown Error: %x", err);
	
	wchar_t szMsg[512];
	wsprintf(szMsg, L"%s failed: %s", psz, szErr);

	char szBuffer[512];
	WideCharToMultiByte( CP_ACP, 0, szMsg, -1, szBuffer, 512, NULL, NULL );
	
	size_t eos = strlen(szBuffer) - 1;
	if (szBuffer[eos-1] = '\r') szBuffer[eos-1] = 0;
	else if (szBuffer[eos] = '\n') szBuffer[eos] = 0;

	env->ThrowNew(clsExc, szBuffer);
	//__asm int 3;
}

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_clientInit
  (JNIEnv *env, jobject jobj, jstring mechanisms, jstring protocol, jstring serverName, jobject props, jobject cbh)
{

	Model* s_pModel = new Model(env, false, _err);
	env->SetLongField(jobj, fidModel, (long long)s_pModel);

	jboolean iscopy;
	const jchar *mechs = env->GetStringChars(mechanisms, &iscopy);
	s_pModel->chooseSPNEGO(mechs);
	env->ReleaseStringChars(mechanisms, mechs);

	s_pModel->chooseCredentials(env);
	if (env->ExceptionCheck()) return;

	s_pModel->chooseContextRequirements(ISC_REQ_IDENTIFY);

}

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_serverInit
  (JNIEnv *env, jobject jobj, jstring mechanism, jstring protocol, jstring serverName, jobject props, jobject cbh)
{

	Model* s_pModel = new Model(env, true, _err);
	env->SetLongField(jobj, fidModel, (long long)s_pModel);

	jboolean iscopy;
	const jchar *mech = env->GetStringChars(mechanism, &iscopy);

	s_pModel->chooseSpecificSSP(mech);

	env->ReleaseStringChars(mechanism, mech);

	s_pModel->chooseCredentials(env);
	if (env->ExceptionCheck()) return;

}

jobjectArray getMechanismNames
  (JNIEnv *env, jclass jobj, bool isServer, jobject map)
{
	Model* s_pModel = new Model(env, isServer, _err);

	ULONG cSSPs;
	PSecPkgInfo pSSPs;
	s_pModel->enumSSPs(env, &cSSPs, &pSSPs);
	if (env->ExceptionCheck()) {
		s_pModel->freeContextBuffer(pSSPs);
		return NULL;
	}

	int cMechs = 0;
	for (ULONG i = 0; i < cSSPs; ++i) {
		if (pSSPs[i].fCapabilities & SECPKG_FLAG_CONNECTION && 
			pSSPs[i].wRPCID != RPC_C_AUTHN_GSS_NEGOTIATE)
			cMechs++;
	}
	jobjectArray mechNames = env->NewObjectArray(cMechs,
		env->FindClass("java/lang/String"),
		env->NewStringUTF(""));

	cMechs = 0;
	for (ULONG i = 0; i < cSSPs; ++i) {
		if (pSSPs[i].fCapabilities & SECPKG_FLAG_CONNECTION && 
			pSSPs[i].wRPCID != RPC_C_AUTHN_GSS_NEGOTIATE)
			env->SetObjectArrayElement(
			mechNames,cMechs++,env->NewString(pSSPs[i].Name, (jsize)wcslen(pSSPs[i].Name)));
	}

	s_pModel->freeContextBuffer(pSSPs);

	delete s_pModel;

	return mechNames;
}

JNIEXPORT jobjectArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_getClientMechanismNames
  (JNIEnv *env, jclass jobj, jobject map)
{
	return getMechanismNames(env, jobj, false, map);
}

JNIEXPORT jobjectArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_getServerMechanismNames
  (JNIEnv *env, jclass jobj, jobject map)
{
	return getMechanismNames(env, jobj, true, map);
}

// Client
JNIEXPORT jbyteArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_evaluateChallenge
  (JNIEnv *env, jobject jobj, jbyteArray challenge)
{
	Model *s_pModel = (Model*)env->GetLongField(jobj, fidModel);

	const int bufsize = 4096 * 4;
	BYTE	s_msgBuf[bufsize];	// a big buffer eliminates buffer management worries
	DWORD	s_cbMsgBuf;			// length of message + security trailer (signature/encryption info)

	jsize size = env->GetArrayLength(challenge);
	jbyte *abyte = env->GetByteArrayElements(challenge, false);

	for (int i = 0; i < size && i < bufsize; i++)
		s_msgBuf[i] = abyte[i];
	env->ReleaseByteArrayElements(challenge,abyte,0);
	s_cbMsgBuf = size;

	void* pOutput;
	DWORD cbOutput = 0;
	bool bContinueNeeded;
	ULONG grfCtxAttrs;

	wchar_t szSPN[256];

	jboolean iscopy;
	jstring authorizationId = (jstring)env->GetObjectField(jobj, fidAuthorizationId);
	const jchar *spn = env->GetStringChars(authorizationId, &iscopy);
	wcsncpy(szSPN, spn, 255);
	env->ReleaseStringChars(authorizationId, spn);

	s_pModel->initializeSecurityContext(env, szSPN,
		s_msgBuf, s_cbMsgBuf,
		&pOutput, &cbOutput,
		&bContinueNeeded, &grfCtxAttrs);
	if (env->ExceptionCheck()) return NULL;

	jbyteArray response = env->NewByteArray(cbOutput);
	abyte = env->GetByteArrayElements(response, false);
	for (DWORD i = 0; i < cbOutput && i < bufsize; i++)
		abyte[i] = ((byte*)pOutput)[i];
	env->ReleaseByteArrayElements(response,abyte,0);

	if (!bContinueNeeded) {
		env->SetBooleanField(jobj, fidCompleted, 1);
	}

	return response;
}

// Server
JNIEXPORT jbyteArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_evaluateResponse
  (JNIEnv *env, jobject jobj, jbyteArray response)
{
	Model *s_pModel = (Model*)env->GetLongField(jobj, fidModel);

	const int bufsize = 4096 * 4;
	BYTE	s_msgBuf[bufsize];	// a big buffer eliminates buffer management worries
	DWORD	s_cbMsgBuf;			// length of message + security trailer (signature/encryption info)

	jsize size = env->GetArrayLength(response);
	jbyte *abyte = env->GetByteArrayElements(response, false);

	for (int i = 0; i < size && i < bufsize; i++)
		s_msgBuf[i] = abyte[i];
	env->ReleaseByteArrayElements(response,abyte,0);
	s_cbMsgBuf = size;

	void* pOutput;
	DWORD cbOutput = 0;
	bool bContinueNeeded;
	ULONG grfCtxAttrs;

	s_pModel->acceptSecurityContext(env, 
		s_msgBuf, s_cbMsgBuf,
		&pOutput, &cbOutput,
		&bContinueNeeded, &grfCtxAttrs);
	if (env->ExceptionCheck()) return NULL;
	
	jbyteArray challenge = env->NewByteArray(cbOutput);
	abyte = env->GetByteArrayElements(challenge, false);
	for (DWORD i = 0; i < cbOutput && i < bufsize; i++)
		abyte[i] = ((byte*)pOutput)[i];
	env->ReleaseByteArrayElements(challenge,abyte,0);
	
	if (!bContinueNeeded) {
		env->SetBooleanField(jobj, fidCompleted, 1);
	}

	return challenge;
}

JNIEXPORT jstring JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_getAuthorizationID
  (JNIEnv *env, jobject jobj)
{
	Model *s_pModel = (Model*)env->GetLongField(jobj, fidModel);

	wchar_t *pvSpn;
	s_pModel->queryContextAttributes(env, SECPKG_ATTR_NAMES, &pvSpn);
	if (env->ExceptionCheck()) {
		s_pModel->freeContextBuffer(pvSpn);
		return NULL;
	}
	jstring names = env->NewString(pvSpn, (jsize)wcslen(pvSpn));
	s_pModel->freeContextBuffer(pvSpn);
	return names;
}
